// Copyright 2012 Google Inc. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.google.api.ads.common.lib.factory;

import com.google.api.ads.common.lib.conf.ConfigurationHelper;
import com.google.api.ads.common.lib.exception.ValidationException;
import com.google.common.annotations.VisibleForTesting;

import org.apache.commons.configuration.Configuration;

import java.io.File;
import java.net.URL;

/**
 * Configuration for Ads*Services classes.
 *
 * @author Adam Rogal
 * @author Joseph DiLallo
 */
public class FactoryConfiguration {

  private Class<?> adWordsSoapModuleClass;
  private Class<?> dfaSoapModuleClass;
  private Class<?> dfpSoapModuleClass;

  /**
   * Private constructor.
   */
  private FactoryConfiguration() {
  }

  /**
   * Gets the SOAP module to use for AdWords.
   */
  public Class<?> getAdWordsSoapModule() {
    return adWordsSoapModuleClass;
  }

  /**
   * Gets the SOAP module to use for DFA.
   */
  public Class<?> getDfaSoapModule() {
    return dfaSoapModuleClass;
  }

  /**
   * Gets the SOAP module to use for DFP.
   */
  public Class<?> getDfpSoapModule() {
    return dfpSoapModuleClass;
  }

  /**
   * Builder for FactoryConfiguration.
   */
  public static class Builder implements
      com.google.api.ads.common.lib.utils.Builder<FactoryConfiguration> {

    private FactoryConfiguration factoryConfiguration;
    private String adWordsSoapModuleClassString;
    private String dfaSoapModuleClassString;
    private String dfpSoapModuleClassString;
    @VisibleForTesting
    ConfigurationHelper configHelper;

    /**
     * Constructor.
     */
    public Builder() {
      this.factoryConfiguration = new FactoryConfiguration();
      this.configHelper = new ConfigurationHelper();
    }

    @VisibleForTesting
    Builder(ConfigurationHelper configHelper) {
      this.factoryConfiguration = new FactoryConfiguration();
      this.configHelper = configHelper;
    }

    /*
     * (non-Javadoc)
     * @see com.google.api.ads.common.lib.utils.Builder#fromFile()
     */
    public Builder fromFile() {
      return fromFile(Builder.DEFAULT_CONFIGURATION_FILENAME);
    }

    /* (non-Javadoc)
     * @see com.google.api.ads.common.lib.utils.Builder#fromFile(java.lang.String)
     */
    public Builder fromFile(String path) {
      return from(configHelper.fromFile(path));
    }

    /* (non-Javadoc)
     * @see com.google.api.ads.common.lib.utils.Builder#fromFile(java.io.File)
     */
    public Builder fromFile(File path) {
      return from(configHelper.fromFile(path));
    }

    /* (non-Javadoc)
     * @see com.google.api.ads.common.lib.utils.Builder#fromFile(java.net.URL)
     */
    public Builder fromFile(URL path) {
      return from(configHelper.fromFile(path));
    }

    /**
     * Reads properties from the provided {@link Configuration} object.<br><br>
     * Known properties:
     * <ul>
     * <li>api.ads.soapClientHandlerModule</li>
     * </ul>
     *
     * @param config
     * @return Builder populated from the Configuration
     */
    public Builder from(Configuration config) {
      String adWordsSoapModuleClassString = config.getString("api.adwords.soapModule", null);
      if (adWordsSoapModuleClassString != null) {
        this.adWordsSoapModuleClassString = adWordsSoapModuleClassString;
      }
      String dfaSoapModuleClassString = config.getString("api.dfa.soapModule", null);
      if (dfaSoapModuleClassString != null) {
        this.dfaSoapModuleClassString = dfaSoapModuleClassString;
      }
      String dfpSoapModuleClassString = config.getString("api.dfp.soapModule", null);
      if (dfpSoapModuleClassString != null) {
        this.dfpSoapModuleClassString = dfpSoapModuleClassString;
      }
      return this;
    }

    /**
     * Specifies which SOAP client handler module class to use for AdWords.
     */
    public Builder withAdWordsSoapModuleClass(Class<?> soapModuleClass) {
      factoryConfiguration.adWordsSoapModuleClass = soapModuleClass;
      return this;
    }

    /**
     * Specifies which SOAP client handler module class to use for DFA.
     */
    public Builder withDfaSoapModuleClass(Class<?> soapModuleClass) {
      factoryConfiguration.dfaSoapModuleClass = soapModuleClass;
      return this;
    }

    /**
     * Specifies which SOAP client handler module class to use for DFP.
     */
    public Builder withDfpSoapModuleClass(Class<?> soapModuleClass) {
      factoryConfiguration.dfpSoapModuleClass = soapModuleClass;
      return this;
    }

    /**
     * Builds the {@code FactoryConfiguration}.
     *
     * @return the built {@code FactoryConfiguration}
     * @throws ValidationException if the {@code FactoryConfiguration} did not
     *         validate
     */
    public FactoryConfiguration build() throws ValidationException {
      from(configHelper.fromSystem());
      validate(factoryConfiguration);
      return factoryConfiguration;
    }

    /**
     * Validates the factory configuration.
     */
    private void validate(FactoryConfiguration factoryConfiguration) throws ValidationException {
      if (adWordsSoapModuleClassString != null) {
        try {
          factoryConfiguration.adWordsSoapModuleClass =
              Class.forName(adWordsSoapModuleClassString);
        } catch (ClassNotFoundException e) {
          throw new ValidationException("Cannot find SOAP module class: "
              + adWordsSoapModuleClassString, "api.adwords.soapModule", e);
        }
      }
      if (dfaSoapModuleClassString != null) {
        try {
          factoryConfiguration.dfaSoapModuleClass =
              Class.forName(dfaSoapModuleClassString);
        } catch (ClassNotFoundException e) {
          throw new ValidationException("Cannot find SOAP module class: "
              + dfaSoapModuleClassString, "api.dfa.soapModule", e);
        }
      }
      if (dfpSoapModuleClassString != null) {
        try {
          factoryConfiguration.dfpSoapModuleClass =
              Class.forName(dfpSoapModuleClassString);
        } catch (ClassNotFoundException e) {
          throw new ValidationException("Cannot find SOAP module class: "
              + dfpSoapModuleClassString, "api.dfp.soapModule", e);
        }
      }
    }
  }
}
